/*----------------------
  TestPlayMidiLib.c

  Test of PlayMidiLib.c

  http://www.tunesmithy.co.uk/playmidilib/index.htm

  E-mail support@tunesmithy.co.uk

  (c) Robert Walker 2002
----------------------*/#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "PlayMidiLib.h"
#include <conio.h>
#include <ctype.h>
#include <math.h>
#define TEN_K 10240
#include "QTMusicLib.h"

#define chord_test
#define harmonic_series_test
#define pc_keyboard_test
#define non_melodic_percussion_test
#define scale_test

typedef int                 BOOL;

#ifndef FALSE
#define FALSE               0
#endif

#ifndef TRUE
#define TRUE                1
#endif

void RemoveLeadingAndTrailingWhiteSpace(char *sz)
{
 int i=0,ipos=0;
 BOOL bWhiteSpaceSoFar=TRUE;
 for(i=strlen(sz);i>=0;i--)
 {
  char c=sz[i];
  if(!isspace(c))
   bWhiteSpaceSoFar=FALSE;
  // remove trailing white spaces
  if(bWhiteSpaceSoFar)
   sz[i]='\0';
 }
 bWhiteSpaceSoFar=TRUE;
 for(i=0;sz[i];i++)
 {
  char c=sz[i];
  if(!isspace(c))
   bWhiteSpaceSoFar=FALSE;
  if(!bWhiteSpaceSoFar)
   sz[ipos++]=c;
 }
 sz[ipos]='\0';
}


static int substrcmpi(const char * sz,const char * szsubstr)
//returns 0 if the same, or szsubstr a sub string of sz
{
 long i=0;
 if(sz==NULL||szsubstr==NULL)                          
  return 0;
 for(i=0;;i++)
 {
  char c1=sz[i];
  char c2=szsubstr[i];
  if(c2=='\0')
   return 0;
  toupper(c1); 
  toupper(c2);
  if(c1!=c2)
   return i+1;
 }
}


BOOL bNewVoice(char *sz)
{
 int i=0;
#define MAX_szVoiceFound 100
 char szVoiceFound[MAX_szVoiceFound];
 if(substrcmpi(sz,"voice")==0)
 {
  int iVoice=FindMidiVoice(&sz[strlen("voice")],szVoiceFound,MAX_szVoiceFound);
  SetMidiVoice(iVoice);
  printf("\n******new voice %s (%d)******\n",szVoiceFound,iVoice);
  return TRUE;
 }
 return FALSE;
}


BOOL bNew1o1(char *sz)
{
 int i=0;
 for(i=strlen(sz);i>=0;i--)
 {
  if(strcmpi(&sz[i],"hz")==0)
  {
   double d1o1=atof(sz);
   printf("\n******new 1/1 %g Hz******\n",d1o1);
   Set1o1(d1o1);
   return TRUE;
  }
 }
 return FALSE;
}

BOOL bProcessInput(char *sz)
{
 RemoveLeadingAndTrailingWhiteSpace(sz);
 if(strcmpi(sz,"voices")==0)
 {
  printf("\n\n******Complete list of voices******\n");
  PrintVoices();
  return TRUE;
 }
 if(bNewVoice(sz))
  return TRUE;
 if(bNew1o1(sz))
  return TRUE;
 return FALSE;
}

BOOL bDoMidiTest(BOOL *pbShowPitchBendRangeTest)
{
 char szInput[TEN_K];
 printf("\n\n******Pitch bend range test******\n\n");
 printf("You should now hear two notes played.\n\n");
 TestPitchBendRange();
 printf
  (
  "What did you hear?\n\n"
  "If you hear the same pitch twice, pitch bends are played as desired, answer Y.\n\n"
  "Otherwise, answer N to see more information and then exit.\n\n"
  "Keep going (Y or N)?\n"
  );
 gets(szInput);
 RemoveLeadingAndTrailingWhiteSpace(szInput);
 if(toupper(szInput[0])=='N')
 {
  printf
  (
   "If you hear two distinct notes a major third apart then no pitch bends are\n"
   "played. You may get other intervals depending on the pitch bend range.\n\n"
   "See if you can adjust the pitch bend range at the sound card or synth.\n"
   "You want a range of plus or minus two semitones. Alternatively,\n"
   "use QuicktTime, or install a soft synth such as the Roland or\n"
   "Yamaha soft synths.\n"
   "Exit? - just press return to exit, or N if you want to try the test again\n"
   );
  gets(szInput);
  if(toupper(szInput[0])=='N')
  {
   printf("****Type \"test\" to try the test again***");
   *pbShowPitchBendRangeTest=TRUE;
   return TRUE;
  }
  return FALSE;
 }
 *pbShowPitchBendRangeTest=FALSE;
 return TRUE;
}

BYTE byRows[8];
BYTE GetIthBit(int ipos)
{
 // returns 1 or 0 depending on the value of the ith bit.
 int irow=ipos/8;
 int ipos_row=ipos-8*irow;
 BYTE byBitVal=0;
 BYTE byK=0;
 int i=0;
 byK=byRows[irow];
 // divide by 2 repeatedly ipos_row times
 for(i=0;i<ipos_row;i++)
  byK/=2;
 // now lowest bit of byK is the one we want to find
 // so now divide by two once more, and then multiply
 // by two to set that bit to 0, and subtract 
 // the result from byK.
 byK=byK-(byK/2)*2;
 return byK; 
}

void SetIthBit(int ipos, BYTE byNewBit)
{
 // sets the ith bit to byNewBit 
 // where byNewBit should be 1 or 0
 BYTE byPrev=GetIthBit(ipos);
 if(byPrev==byNewBit)
  // nothing to do
  return;
 {
  int irow=ipos/8;
  int ipos_row=ipos-8*irow;
  BYTE byBit=1;
  int i=0;
  // multiply by 2 repeatedly to get the
  // bit in the right position
  for(i=0;i<ipos_row;i++)
   byBit*=2;
  // now byBit is the desired new ith bit
  // so we can simply add it or subtract it
  // as we know now that it differs from
  // the previous value for the ith bit
  if(byNewBit==0)
   byRows[irow]-=byBit;
  else
   byRows[irow]+=byBit;
 }
}


int main()
{
 int i=0;
 char szInput[TEN_K];
 BOOL bQTAvailable=FALSE;
 int nOutDevs=nNumOutDevices();
 int iDevSel=0;
#define MAX_szDeviceName 1024
 char szDeviceName[MAX_szDeviceName];
 BOOL bShowPitchBendRangeTest=TRUE;
 for(i=0;i<64;i++)
 {
  SetIthBit(i,1);
 }
 for(i=0;i<64;i++)
 {
  BYTE bit=GetIthBit(i);
  if(bit!=1)
   bit=bit;
 }
 for(i=0;i<64;i++)
 {
  BYTE bit=0;
  SetIthBit(i,0);
 }
 for(i=0;i<64;i++)
 {
  BYTE bit=GetIthBit(i);
  if(bit!=0)
   bit=bit;
 }
 printf("Test of the PlayMidiLib routines.\n\n");
 start:;
 printf("Choose your Midi Out device to use - type its letter to select\n"
        "Or just press return to select the first one.\n\n");
 for(i=0;i<nOutDevs;i++)
  printf("%c - %s\n",'A'+i,szNameForDevice(i,szDeviceName,MAX_szDeviceName));
 bQTAvailable=CheckForQTLibrary();
 if(!bQTAvailable)
  printf("%c - How to install QuickTime for Midi apps to use\n",'A'+nOutDevs);
 gets(szInput);
 RemoveLeadingAndTrailingWhiteSpace(szInput);
 {
  char cdev=szInput[0]=='\0'?'A':toupper(szInput[0]);
  iDevSel=cdev-'A';
  if(iDevSel==nOutDevs&&!bQTAvailable)
  {
   printf(SZ_QT_INSTALL_INSTRUCTIONS);
   printf
   (
   "Press return to exit."
   );
   getchar();
   exit(0);
  }
  printf("Device selected:\n");
  printf("%c - %s\n",'A'+iDevSel,szNameForDevice(iDevSel,szDeviceName,MAX_szDeviceName));
  szNameForDevice(iDevSel,szDeviceName,MAX_szDeviceName);
  if(strcmpi(szDeviceName,"QuickTime")==0)
   printf("\n\nInitialising QuickTime - may take a little while\n");
 }
 if(StartMidiOutForiDev(iDevSel)==MIDI_OUT_ERROR)
 {
  if(strcmpi(szDeviceName,"QuickTime")==0)
  {
   printf
   ("\n\nCouldn't start up Quicktime\n\n"
   );
   printf
   (
    SZ_QT_INSTALL_INSTRUCTIONS
    );
   printf
   (
   "Press return to exit."
   );
  }
  else
  printf
   ("Couldn't open Midi Out - perhaps it is already in use for another application.\n\n"
    "Press return to exit."
   );
  getchar();
  exit(0);
 }
 if(strcmpi(szDeviceName,"QuickTime")==0)
  printf("QT initialised okay\n");
 SetMidiVoice(6);//Harpshichord
#ifdef chord_test
 for(;;)
 {
  int nChordNotes=0;
  printf
  (
  "\n\n****Things you can do at any time****\n\n"
  "To change pitch of 1/1 enter frequency followed by \"hz\"\n"
  "You can use \"0 hz\" to reset to middle c\n"
  "To change voice, enter \"voice\" then the instrument name\n"
  "e.g. \"voice harpsichord\" , to change voice to harpsichord,\n"
  "You can use just part of the instrument name, or it's number.\n"
  "For complete list of voices type \"voices\""
  "\n\n****Chord test****\n\n"
  "Enter a chord to play in cents or ratios notation as intervals from 1/1\n"
  "Include decimal point for cents (SCALA convention).\n\n"
  "Just press return for the preset of a septimal minor chord 7/6 3/2 2/1\n\n"
  "Type \"next\" to move on to next option, \"start\" to start again or \"exit\" to exit.%s\n"
  ,bShowPitchBendRangeTest
  ?"\n\nRecommended first time you try a device:\n"
  "Type \"test\" to test the pitch bend range of your sound card / synth."
  :""
  );
  bShowPitchBendRangeTest=FALSE;
#define PROCESS_NEXT_INPUT(szInput)\
  gets(szInput);\
  if(bProcessInput(szInput))\
   continue;\
  if(strcmpi(szInput,"test")==0)\
  {\
   if(!bDoMidiTest(&bShowPitchBendRangeTest))\
    goto do_exit;\
   continue;\
  }\
  if(strcmpi(szInput,"start")==0)\
   goto start;\
  if(strcmpi(szInput,"exit")==0)\
   goto do_exit;
  PROCESS_NEXT_INPUT(szInput)
  if(strcmpi(szInput,"next")==0)
   break;\
  if(szInput[0]=='\0')
   strcpy(szInput,"7/6 3/2 2/1");
  printf("You should now hear your chord as a broken chord\n\n");
  SetMidiScale(szInput);
  SleepSeconds(0.5);//best to sleep here - e.g. QT needs time to apply the pitch bends
  nChordNotes=GetnScaleNotes();
  for(i=0;i<nChordNotes;i++)
  {
   PlayScaleDegree(i,127);
   SleepSeconds(1);
  }
  PrintNotesInPlay();
  SleepSeconds(1);
  for(i=0;i<nChordNotes;i++)
   StopScaleDegree(i,127);
 }
#endif
#ifdef scale_test
 SetMidiScale("9/8 5/4 4/3 3/2 5/3 15/8 2/1");
 {
  BOOL bFirstTime=TRUE;
  for(;;)
  {
   int nScaleNotes=0;
   printf
   ("\n\n****Scale test****\n\n"
   "Enter a scale to play in cents or ratios notation as intervals from 1/1\n"
   "Include decimal point for cents (SCALA convention).\n\n"
   "You can use \"n(1/7)\" for 7 equal steps to an octave.\n\n" 
   "Also use \"= 7\" for 7 equal steps to an octave.\n\n" 
   "You can also use:\n"
   "h 2 for harmonic series.\n\n"
   );
   if(bFirstTime)
   printf
   ("Just press return for preset of the just intonation diatonic\n"
   "9/8 5/4 4/3 3/2 5/3 15/8 2/1\n"
   "\"next\" to move on to next option, \"start\" to start again or \"exit\" to exit.%s\n"
   );
   else
   printf
   ("Just press return to keep the current scale:\n"
   "%s\n"
   "\"next\" to move on to next option, \"start\" to start again or \"exit\" to exit.%s\n"
   ,GetszMostRecentScale()
   );
   PROCESS_NEXT_INPUT(szInput)
   if(strcmpi(szInput,"next")==0)
    break;\
   bFirstTime=FALSE;
   if(szInput[0]!='\0')
   {
    if(szInput[0]=='=')
     // make it a ratio with the /1 as otherwise it will get read as cents if it has a decimal point in it
     sprintf(szInput,"n(%g/1)",1/max(1e-10,atof(&szInput[1])));
    SetMidiScale(szInput);
   }
   nScaleNotes=GetnScaleNotes();
   printf
    ("Enter the number of notes of the scale to play (-ve number to descend from 1/1)\n"
    "Or press return to play %d notes\n",nScaleNotes);
   if(szInput[0]!='\0')
    gets(szInput);
   nScaleNotes=atoi(szInput);
   printf("You should now hear your ascending scale of %d notes\n\n",nScaleNotes);
   SleepSeconds(0.5);//best to sleep here - e.g. QT needs time to apply the pitch bends
   {
    int iincr=nScaleNotes<0?-1:1;
    for(i=0;i!=nScaleNotes;i+=iincr)
    {
     PlayScaleDegree(i,127);
     SleepSeconds(0.5);
     StopScaleDegree(i,127);
    }
   }
  }
 }
#endif
#ifdef harmonic_series_test
 bNew1o1("55.0 hz");
 {
  int nChordNotes=10;
  int nChordsToPlay=100;
  double dtstart=0,dtdesired=0;
  for(;;)
  {
   double dVelocity=127;
   double dtime=1;
   printf
   ("\n\n****Harmonic Series test****\n\n"
    "Now, a drifting harmonic series chord, drifting up and then down\n"
    "the harmonic series.\n"
    "Enter the number of notes in the chord.\n"
    "To change the number of chords to play enter \"n 50\" to play 50 chords\n"
    "To change pitch of 1/1 enter frequency followed by \"hz\", or \"0 hz\" to reset to middle c\n"
    "Press return only for preset values of 10 notes and 100 chords\n"
    "\"next\" to move on to next option, \"start\" to start again or \"exit\" to exit.%s\n"
   );
   PROCESS_NEXT_INPUT(szInput)
   if(strcmpi(szInput,"next")==0)
    break;\
   {
    int i=0;
    for(i=0;szInput[i];i++)
    if(szInput[i]=='n')
    {
     nChordsToPlay=atoi(&szInput[i+1]);
     if(nChordsToPlay==0)
      nChordsToPlay=100;
     printf("\n****Play %d chords****\n",nChordsToPlay);
     continue;
    }
   }
   if(szInput[0]!='\0')
    nChordNotes=atoi(szInput);
   if(nChordNotes==0)
    nChordNotes=20;
   SetMidiScale("h 2");
   dtime=1;
   dVelocity=127.0;
   printf("\nDrifting harmonic series of %d chords, %d notes per chord\n",nChordsToPlay,nChordNotes);
   for(i=0;i<nChannelsAvailable();i++)
    DoBendForScaleDegreeInLoop(i);
   SleepSeconds(0.5);
   dtstart=dSecsSinceStartMidiOut();
   dtdesired=dtstart;
   for(i=0;i<nChordsToPlay+nChordNotes;i++)
   {
    if(i<nChordsToPlay)
    {
     PlayScaleDegree(i,(int)dVelocity);
     dVelocity=127*exp(i*log(0.95));
     dtime=exp(i*log(0.95));
     dtdesired+=dtime;
     SleepSeconds(dtdesired-dSecsSinceStartMidiOut());
    }
    if(i<nChordsToPlay+nChordNotes-16)
     // do the bend well in advance - may get artefacts otherwise
     DoBendForScaleDegree(i+20);
    if(i>=nChordNotes)
    {
     StopScaleDegree(i-nChordNotes,127);
    }
   }
   for(i=0;i<nChannelsAvailable();i++)
    DoBendForScaleDegreeInLoop(nChordsToPlay+nChordNotes-1-i);
   for(i=nChordsToPlay+nChordNotes-1;i>=0;i--)
   {
    if(i>=nChordNotes)
    {
     PlayScaleDegree(i-nChordNotes,(int)dVelocity);
     dVelocity=127*exp(i*log(0.95));
     dtime=exp(i*log(0.95));
     dtdesired+=dtime;
     SleepSeconds(dtdesired-dSecsSinceStartMidiOut());
    }
    if(i>=nChannelsAvailable())
     // do the bend well in advance - may get artefacts otherwise
     DoBendForScaleDegree(i-nChannelsAvailable());
    if(i<nChordsToPlay)
    {
     StopScaleDegree(i,127);
    }
    SleepSeconds(dtime);
   }
  }
 }
#endif
#ifdef pc_keyboard_test
 bNew1o1("0.0 hz");
 for(;;)
 {
  int nChordNotes=0;
  int nChordsToPlay=300;
  int iDegree=-1;
  double dtime=1;
  BOOL bSustain=FALSE;
  StopAllNotesPlayed(127);
  printf
  ("\n\n****Play from p.c. keyboard test****\n\n"
  "Now, enter a scale to play in from p.c. keyboard\n"
  "Scale repeats at last interval - end with 2/1 for octave scales.\n\n"
  "Just press return for the preset of the just intonation diatonic\n"
  "9/8 5/4 4/3 3/2 5/3 15/8 2/1\n"
  "To change pitch of 1/1 enter frequency followed by \"hz\", or \"0 hz\" to reset to middle c\n"
  "\"next\" to move on to next option, \"start\" to start again or \"exit\" to exit.%s\n"
  );
  PROCESS_NEXT_INPUT(szInput)
  if(strcmpi(szInput,"next")==0)
   break;\
  SetMidiScale(szInput);
  SleepSeconds(0.5);
  if(szInput[0]=='\0')
   strcpy(szInput,"9/8 5/4 4/3 3/2 5/3 15/8 2/1");
  else
  if(bProcessInput(szInput))
   continue;
  if(szInput[0]=='.')
   SetMidiScale(&szInput[1]);
  else
   SetMidiScale(szInput);
  SleepSeconds(0.5);
  printf
  ("Now, play in the scale using numbers 0 to 9, or letters a to z for the\n"
  "scale degree. Press spaces to switch sustain on / off\n"
  "To change scale or pitch of 1/1 press return key.\n"
  );
  for(;;)
  {
   char c=_getche();
   switch(c)
   {
    case '\r':
    case '\n':
     getchar();
     goto next_scale;
    case ' ':
     bSustain=bSustain?FALSE:TRUE;
     if(!bSustain)
      StopAllNotesPlayed(127);
     continue;//toggle sustain
   }
   if(iDegree>=0&&!bSustain)
    StopScaleDegree(iDegree,127);
   if(c>='0'&&c<='9')
    iDegree=c-'0';
   else
   if(c>='a'&&c<='z')
    iDegree=10+c-'a';
   else
   if(c>='A'&&c<='Z')
    iDegree=10+c-'A';
   else
   switch(c)
   {
   case '`':
   case '':
   case '\\':
   case '|':iDegree=37;break;
   case '-':
   case '_':iDegree=37;break;
   case '=':
   case '+':iDegree=38;break;
   case '[':
   case '{':iDegree=39;break;
   case ']':
   case '}':iDegree=40;break;
   case ';':
   case ':':iDegree=41;break;
   case '\'':
   case '@':iDegree=42;break;
   case '#':
   case '~':iDegree=43;break;
   case '<':
   case ',':iDegree=44;break;
   case '>':
   case '.':iDegree=45;break;
   case '?':
   case '/':iDegree=46;break;
   }
   if(iDegree>=0)
    PlayScaleDegree(iDegree,127);
  }
  next_scale:;
 } 
#endif
#ifdef non_melodic_percussion_test
 {
  int iDegree=-1;
  BOOL bSustain=FALSE;
  StopAllNotesPlayed(127);
  printf
  ("\n\n****Non Melodic Percussion test****\n\n"
  "Use numbers 0 to 9, letters a to z or punctuation\n"
  "to play various non melodic percussion instruments.\n\n"
  "Press spaces to switch sustain on / off, and press return key to exit program.\n"
  );
  bSustain=FALSE;
  for(;;)
  {
   char c=_getche();
   switch(c)
   {
    case '\r':
    case '\n':
     getchar();
     StopAllNotesPlayed(127);
     printf("\nExit Program (press return again to exit, or N to continue)\n");
     gets(szInput);
     RemoveLeadingAndTrailingWhiteSpace(szInput);
     c=szInput[0];
     if(toupper(c)!='N')
      goto do_exit;
     break;
    case ' ':
     bSustain=bSustain?FALSE:TRUE;
     if(!bSustain)
      StopAllNotesPlayed(127);
     continue;//toggle sustain
   }
   if(iDegree>=0&&!bSustain)
    StopNonMelodicPercussion(FIRST_GM_NON_MELOD_PERC+iDegree,127);
   if(c>='0'&&c<='9')
    iDegree=c-'0';
   else
   if(c>='a'&&c<='z')
    iDegree=10+c-'a';
   else
   if(c>='A'&&c<='Z')
    iDegree=10+c-'A';
   else
   switch(c)
   {
   case '-':
   case '_':iDegree=36;break;
   case '+':
   case '=':iDegree=37;break;
   case '{':
   case '[':iDegree=38;break;
   case '}':
   case ']':iDegree=39;break;
   case ':':
   case ';':iDegree=40;break;
   case '\'':
   case '@':iDegree=41;break;
   case '~':
   case '#':iDegree=42;break;
   case ',':
   case '<':iDegree=43;break;
   case '.':
   case '>':iDegree=44;break;
   case '?':
   case '/':iDegree=45;break;
   case '':
   case '`':iDegree=46;break;
   case '|':
   case '\\':iDegree=46;break;
   }
   if(iDegree>=0)
    PlayNonMelodicPercussion(FIRST_GM_NON_MELOD_PERC+iDegree,127);
  }
 }
#endif
 do_exit:;
 StopMidiOut();
 return 0;
}
